#include <stdio.h>
#include <string.h>

// Definizione "leggera" dei booleani
// presa dalla Standard Template Library

#define bool int
#define true 1
#define false 0


// Nuova definizione della funzione strdup, quella standard 
// della libreria C e' basata sulla malloc, questa, piu'
// correttamente sugli operatori new e delete del C++
char *strdupnew(char const *str)    {
        return (strcpy(new char [strlen(str) + 1], str));    
}

/********************************************************************
 Classe Astratta Node, 
 pensata per rappresentare gli oggetti da inserire nell'albero. 
 Come tutte le classi astratte va usata come base class per
 derivare la classe che poi si utilizzera' effettivamente nell'albero.
********************************************************************/

class Node
{
public:

virtual ~Node();	// destructor

// duplicator
// funzione pure virtual che crea una copia del nodo 
virtual Node* duplicate() const = 0;	

// comparison of 2 objects
virtual int compare(Node const *other) const = 0; 

// function to do whatever is needed to the node
virtual void process() = 0;

// called when object to add was already in the tree
virtual void already_stored();
};

Node::~Node()
{}

void Node::already_stored()
{}

/********************************************************************
Classe concreta derivata dalla classe Node che implementa
l'oggetto che utilizzeremo per contare le occorrenze di ogni
parola in un file.

********************************************************************/
class Strnode: public Node    {
public: // destructor, constructors
	~Strnode();
	Strnode();            
	Strnode(Strnode const &other); // copy constructor
	Strnode(char const *s);	
	
	// assignment
	Strnode const &operator=(Strnode const &other);
	
	// functions required by Node protocol 
	// erano state definite virtual nella classe base Node
	Node* duplicate() const;
	int compare(Node const *other) const; 
	void process();
	void already_stored(); 

// data 
private:          
	char *str;        
    int times;   // numero di occorrenze finora trovate
};

Strnode::~Strnode(){ delete str;    }    

Strnode::Strnode() { str = 0; times = 0;    }

Strnode::Strnode(Strnode const &other) {
	str = strdupnew(other.str);        
	times = other.times;
}

Strnode::Strnode(char const *s) {
	str = strdupnew(s);
	times = 1;    
}

Strnode const &Strnode::operator=(Strnode const &other) {
	if (this != &other) {
		delete str;
		str = strdupnew(other.str);     
		times = other.times;       
	}
    return (*this);
}

Node *Strnode::duplicate() const    { return (new Strnode(*this)); }

int Strnode::compare(Node const *other) const {
	Strnode *otherp = (Strnode *) other;    

    if (str && otherp->str)
           return (strcmp(otherp->str, str));
	
	if (! str && ! otherp->str)     
		return (0);
	
	return ((int) otherp->str - (int) str );    
}

// Chiamata durante l'attraversamento dell'albero,
// nel nostro caso si limita a stampare la stringa e il numero di occorrenze
void Strnode::process(){
	if (str) printf("%4d\t%s\n", times, str);   
}

// Chiamata ogni volta che si cerca di memorizzare nell'albero lo stesso nodo
// piu' di una volta, incrementa il contatore delle occorrenze 
void Strnode::already_stored() {
	times++;    
}


// La classe tree  responsabile della gestione degli oggetti derivati da Node.
// La struttura dellalbero e mantenuta con i soliti puntatori 
// a figlio sinistro destro e allinformazioni (Node):

class Tree { 
public: // destructor, constructors 
    ~Tree(); 
    Tree(); 
    Tree(Tree const &other); 
// assignment
    Tree const &operator=(Tree const &other);
    void add(Node *what); 
	bool del(Node *what);  
	
// processing 
    void preorder_walk();
    void inorder_walk(); 
    void postorder_walk(); 
private: 
// primitives 
    void copy(Tree const &other); 
    void destroy(); 
	Node *pop();
	Tree *findfirst() const;
// data 
    Tree *left, *right;
    Node *info; 
};


// La distruzione dell'albero e' effettuata distruggendo
// ricorsivamente tutti gli oggetti che lo compongono.
// si usano due funzioni mutuamente ricorsive.

Tree::~Tree() {
    destroy();
}

void Tree::destroy() {
    delete info; 
    if (left) delete left; 
    if (right) delete right; 
}

// default constructor: initializes to 0
Tree::Tree() {
	left = right = 0;
	info = 0;
}

// copy constructor: inizializza l'oggetto appena creato con il 
// contenuto di un altro oggetto (other)
Tree::Tree(Tree const &other) {
	copy(other);
}

// funzione privata che copia ricorsivamente un albero,
// notare che la copia dei sottoalberi e' effettuata usando il 
// copy constructor definito sopra

void Tree::copy(Tree const &other) {
  info = other.info ? other.info->duplicate() :0; 
  left = other.left ? new Tree(*other.left) :0; 
 right = other.right ? new Tree(*other.right) :0; 
}


// overloading dell'operatore di assegnamento.
// notare il controllo per evitare l'autoassegnamento a=a che 
// altrimenti distruggerebbe l'oggetto
Tree const &Tree::operator=(Tree const &other) {
	if (this != &other)
	{
		destroy();
		copy(other);
	}
	return (*this);
}


// Funzione ricorsiva che aggiunte un nodo all'albero
// Si possono presentare due situazioni:
// - Il campo info del nodo corrente  uno 0-pointer, 
//   allora si aggiunge un duplicato delloggetto al nodo corrente
// - Se lalbero  parzialmente pieno si sceglie se metterlo nel 
//   sottoalbero sinistro o destro a seconda del valore restituito 
//   dalla funzione compare.

void Tree::add(Node *what){
	if (! info) info = what->duplicate();
	else {
        int cmp = info->compare(what);
        if (cmp < 0){
            if (! left){
                left = new Tree;
                left->info = what->duplicate();
            }
            else left->add(what);
		}
		else { 
			if (cmp > 0){
				if (! right){
					right = new Tree;
					right->info = what->duplicate();
				}
				else right->add(what);
			}
			else info->already_stored();
		}
    }
}


// La funzione del(Node *) toglie un elemento dall'albero.
// Prima cerca l'elemento e poi, utilizzando la funzione pop(),
// lo elimina

bool Tree::del(Node *what){
	if (!info) return false;
	int cmp = info->compare(what);
 
	if (cmp < 0){
		if (!left) return false;
		else return left->del(what);
	}
	else { 
		if (cmp > 0){
			if (! right) return false;
			else return right->del(what);
		}
		else {// cmp == 0 we have found the element
			delete pop();
			return true;
		}
	}
}

// Prende un albero ed elimina la radice
// facendo galleggiare il figlio sinistro
// il sottoalbero sinistro e destro del figlio sinistro 
// sono ridistribuiti come di regola 
Node *Tree::pop(){
Node *tmp=info;
	if (left){
		Tree *ll = left->left;
		Tree *lr = left->right;
		Node *li = left->info;
		
		left->left = left->right = 0;
		left->info = 0;
		delete left;
		
		info = li;
		left = ll;
		if (!right) right = lr;
		else (right->findfirst())->left = lr;
	} 
	else if (right) {
		Tree *rl = right->left;
		Tree *rr = right->right;
		Node *ri = right->info;
		
		// attenzione a potare bene l'albero 
		// prima di invocare il distruttore!
		right->left = right->right = 0;
		right->info = 0;
		delete right;

		info= ri;
		right = rr;
		left  = rl;		
	}
	else info = 0;
	return tmp;
}

// Restituisce il primo elemento dell'albero 
// il primo che non ha un sotto albero sinistro

Tree *Tree::findfirst() const {
	if(left) return left->findfirst();
	else return (Tree *) this;
}



// Funzioni di attraversamento dell'albero, per ogni nodo attraversato 
// invocano la funzione process().
// Osservazione sulle classi virtuali:
//
// Se il membro process() della classe Node non fosse stato dichiarato
// come virtual, la chiamata info->process() avrebbe invocato la 
// funzione definita nella classe base Node anziche' quella a cui appartiene
// l'oggetto effettivamente puntato da info (in questo caso un oggetto 
// della classe StrNode)

void Tree::preorder_walk() {
    if (info) info->process();
    if (left)  left->preorder_walk();
    if (right) right->preorder_walk(); 
}

void Tree::inorder_walk() {
    if (left)  left->inorder_walk();
    if (info) info->process();
    if (right) right->inorder_walk(); 
}

void Tree::postorder_walk() {
    if (left)  left->postorder_walk();
    if (right) right->postorder_walk(); 
    if (info) info->process();
}





// Questa funzione utilizza quanto definito per contare le occorrenze di
// ogni parola in un file
//
// - ogni parola letta dal file viene usata per costruire un oggetto 
//   di tipo StrNode che viene aggiunto allalbero,
// - la funzione already_stored()  chiamata automaticamente dalla 
//   classe Tree ogni volta che si aggiunge una parola gi presente nellalbero
// - alla fine del file si attraversa lalbero, stampando in ordine 
// - alfabetico le parole e il numero di occorrenze

void countfile(FILE *inf) {
	Tree  tree;
	char  buf [255];
	while (1)  {
		fscanf(inf, " %s", buf);  
		if (feof(inf)) break;
		Strnode *word = new Strnode(buf);   
		tree.add(word);
		delete word; 
       }  
	tree.inorder_walk();    
}
// Si utilizza il metodo "del", dati due file, per contare le occorrenze 
// delle parole che compaiono nel primo ma non nel secondo
void DiffFile(FILE *inf1, FILE *inf2) {
	Tree  tree;
	char  buf [255];
	while (1)  {
		fscanf(inf1, " %s", buf);  
		if (feof(inf1)) break;
		Strnode *word = new Strnode(buf);   
		tree.add(word);
		delete word; 
       }  
	tree.inorder_walk();    
	printf("\nNow the difference\n");
	while (1)  {
		fscanf(inf2, " %s", buf);  
		if (feof(inf2)) break;
		Strnode *word = new Strnode(buf);   
		if(tree.del(word)) printf("deleted '%s'\n",buf);
		else printf("not deleted '%s'\n",buf);
		delete word; 
       }  
	tree.inorder_walk();    
	}


// main()
// Qualche inizializzazione e controllo (se il file esiste o se si 
// legge da stdin) e poi si invoca la funzione difffile(...) sopra definita.

int main(int argc, char **argv){
	if (argc > 2) {
			FILE *inf1 = fopen(argv [1], "r");
			FILE *inf2 = fopen(argv [2], "r");

			if (! inf1 || ! inf2) {
                    fprintf(stderr, "wordc: can't open file\"%s\"\n", 
						inf1 ? argv [2] : argv [1]);                    
            }
			else DiffFile(inf1,inf2);
		}
	return 1;  
}
